@@ -8,11 +8,18 @@ import base64
 import collections
 from datetime import timedelta, datetime
 from contextlib import contextmanager
-from mock import ANY, patch, Mock
+try:
+    from mock import ANY, patch, Mock
+except ImportError:
+    from unittest.mock import ANY, patch, Mock
 from semantic_version import Version
 
 
 from werkzeug import __version__ as werkzeug_version
+try:
+    from werkzeug.middleware.proxy_fix import ProxyFix
+except ImportError:
+    from werkzeug.contrib.fixers import ProxyFix
 from flask import (
     Flask,
     Blueprint,
@@ -30,7 +37,8 @@ from flask_login import (LoginManager, UserMixin, AnonymousUserMixin,
                          user_needs_refresh, make_next_param, login_url,
                          login_fresh, login_required, session_protected,
                          fresh_login_required, confirm_login, encode_cookie,
-                         decode_cookie, set_login_view, user_accessed)
+                         decode_cookie, set_login_view, user_accessed,
+                         FlaskLoginClient)
 from flask_login.__about__ import (__title__, __description__, __url__,
                                    __version_info__, __version__, __author__,
                                    __author_email__, __maintainer__,
@@ -117,7 +125,7 @@ USERS = {1: notch, 2: steve, 3: creeper, u'佐藤': germanjapanese}
 
 
 class AboutTestCase(unittest.TestCase):
-    """Make sure we can get version and other info."""
+    '''Make sure we can get version and other info.'''
 
     def test_have_about_data(self):
         self.assertTrue(__title__ is not None)
@@ -141,6 +149,10 @@ class StaticTestCase(unittest.TestCase):
         lm = LoginManager()
         lm.init_app(app)
 
+        @lm.user_loader
+        def load_user(user_id):
+            return USERS[int(user_id)]
+
         with app.test_client() as c:
             c.get('/static/favicon.ico')
             self.assertTrue(current_user.is_anonymous)
@@ -152,6 +164,10 @@ class StaticTestCase(unittest.TestCase):
         lm = LoginManager()
         lm.init_app(app)
 
+        @lm.user_loader
+        def load_user(user_id):
+            return USERS[int(user_id)]
+
         with app.test_client() as c:
             with listen_to(user_accessed) as listener:
                 c.get('/static/favicon.ico')
@@ -179,16 +195,18 @@ class InitializationTestCase(unittest.TestCase):
     def test_login_disabled_is_set(self):
         login_manager = LoginManager(self.app, add_context_processor=True)
         self.assertFalse(login_manager._login_disabled)
+        with self.app.app_context():
+            login_manager._login_disabled = True
+            self.assertTrue(login_manager._login_disabled)
 
     def test_no_user_loader_raises(self):
         login_manager = LoginManager(self.app, add_context_processor=True)
         with self.app.test_request_context():
-            session['user_id'] = '2'
+            session['_user_id'] = '2'
             with self.assertRaises(Exception) as cm:
-                login_manager.reload_user()
-            expected_exception_message = 'No user_loader has been installed'
-            self.assertTrue(
-                str(cm.exception).startswith(expected_exception_message))
+                login_manager._load_user()
+            expected_message = 'Missing user_loader or request_loader'
+            self.assertTrue(str(cm.exception).startswith(expected_message))
 
 
 class MethodViewLoginTestCase(unittest.TestCase):
@@ -196,7 +214,7 @@ class MethodViewLoginTestCase(unittest.TestCase):
         self.app = Flask(__name__)
         self.login_manager = LoginManager()
         self.login_manager.init_app(self.app)
-        self.login_manager._login_disabled = False
+        self.app.config['LOGIN_DISABLED'] = False
 
         class SecretEndpoint(MethodView):
             decorators = [
@@ -230,7 +248,7 @@ class LoginTestCase(unittest.TestCase):
         self.app.config['REMEMBER_COOKIE_NAME'] = self.remember_cookie_name
         self.login_manager = LoginManager()
         self.login_manager.init_app(self.app)
-        self.login_manager._login_disabled = False
+        self.app.config['LOGIN_DISABLED'] = False
 
         @self.app.route('/')
         def index():
@@ -318,10 +336,6 @@ class LoginTestCase(unittest.TestCase):
 
         unittest.TestCase.setUp(self)
 
-    def _get_remember_cookie(self, test_client):
-        our_cookies = test_client.cookie_jar._cookies['localhost.local']['/']
-        return our_cookies[self.remember_cookie_name]
-
     def _delete_session(self, c):
         # Helper method to cause the session to be deleted
         # as if the browser was closed. This will remove
@@ -375,7 +389,7 @@ class LoginTestCase(unittest.TestCase):
     def test_login_user_with_header(self):
         user_id = 2
         user_name = USERS[user_id].name
-        self.login_manager.request_callback = None
+        self.login_manager._request_callback = None
         with self.app.test_client() as c:
             basic_fmt = 'Basic {0}'
             decoded = bytes.decode(base64.b64encode(str.encode(str(user_id))))
@@ -386,7 +400,7 @@ class LoginTestCase(unittest.TestCase):
     def test_login_invalid_user_with_header(self):
         user_id = 9000
         user_name = u'Anonymous'
-        self.login_manager.request_callback = None
+        self.login_manager._request_callback = None
         with self.app.test_client() as c:
             basic_fmt = 'Basic {0}'
             decoded = bytes.decode(base64.b64encode(str.encode(str(user_id))))
@@ -429,7 +443,7 @@ class LoginTestCase(unittest.TestCase):
     def test_logout_without_current_user(self):
         with self.app.test_request_context():
             login_user(notch)
-            del session['user_id']
+            del session['_user_id']
             with listen_to(user_logged_out) as listener:
                 logout_user()
                 listener.assert_heard_one(self.app, user=ANY)
@@ -515,6 +529,25 @@ class LoginTestCase(unittest.TestCase):
                              'http://localhost/login')
             self.assertEqual(c.get('/login').data.decode('utf-8'), '/secret')
 
+    def test_unauthorized_with_next_in_strong_session(self):
+        self.login_manager.login_view = 'login'
+        self.app.config['SESSION_PROTECTION'] = 'strong'
+        self.app.config['USE_SESSION_FOR_NEXT'] = True
+
+        @self.app.route('/login')
+        def login():
+            if(current_user.is_authenticated):
+                # Or anything that touches current_user
+                pass
+            return session.pop('next', '')
+
+        with self.app.test_client() as c:
+            result = c.get('/secret')
+            self.assertEqual(result.status_code, 302)
+            self.assertEqual(result.location,
+                             'http://localhost/login')
+            self.assertEqual(c.get('/login').data.decode('utf-8'), '/secret')
+
     def test_unauthorized_uses_blueprint_login_view(self):
         with self.app.app_context():
 
@@ -750,7 +783,7 @@ class LoginTestCase(unittest.TestCase):
 
         with self.assertRaises(Exception) as cm:
             with self.app.test_request_context():
-                session['user_id'] = 2
+                session['_user_id'] = 2
                 self.login_manager._set_cookie(None)
 
         expected_exception_message = 'REMEMBER_COOKIE_DURATION must be a ' \
@@ -849,7 +882,7 @@ class LoginTestCase(unittest.TestCase):
     def test_user_loaded_from_header_fired(self):
         user_id = 1
         user_name = USERS[user_id].name
-        self.login_manager.request_callback = None
+        self.login_manager._request_callback = None
         with self.app.test_client() as c:
             with listen_to(user_loaded_from_header) as listener:
                 headers = [
@@ -998,6 +1031,16 @@ class LoginTestCase(unittest.TestCase):
             # Ensure that if nothing changed the session is not modified.
             self.assertFalse(session.modified)
 
+    def test_invalid_remember_cookie(self):
+        domain = self.app.config['REMEMBER_COOKIE_DOMAIN'] = '.localhost.local'
+        with self.app.test_client() as c:
+            c.get('/login-notch-remember')
+            with c.session_transaction() as sess:
+                sess['_user_id'] = None
+            c.set_cookie(domain, self.remember_cookie_name, 'foo')
+            result = c.get('/username')
+            self.assertEqual(u'Anonymous', result.data.decode('utf-8'))
+
     #
     # Session Protection
     #
@@ -1192,7 +1235,7 @@ class LoginTestCase(unittest.TestCase):
         def protected():
             return u'Access Granted'
 
-        self.app.login_manager._login_disabled = True
+        self.app.config['LOGIN_DISABLED'] = True
 
         with self.app.test_client() as c:
             result = c.get('/protected')
@@ -1246,6 +1289,117 @@ class LoginTestCase(unittest.TestCase):
             self.assertIsInstance(_ucp()['current_user'], AnonymousUserMixin)
 
 
+class LoginViaRequestTestCase(unittest.TestCase):
+    ''' Tests for LoginManager.request_loader.'''
+
+    def setUp(self):
+        self.app = Flask(__name__)
+        self.app.config['SECRET_KEY'] = 'deterministic'
+        self.app.config['SESSION_PROTECTION'] = None
+        self.remember_cookie_name = 'remember'
+        self.app.config['REMEMBER_COOKIE_NAME'] = self.remember_cookie_name
+        self.login_manager = LoginManager()
+        self.login_manager.init_app(self.app)
+        self.app.config['LOGIN_DISABLED'] = False
+
+        @self.app.route('/')
+        def index():
+            return u'Welcome!'
+
+        @self.app.route('/login-notch')
+        def login_notch():
+            return unicode(login_user(notch))
+
+        @self.app.route('/username')
+        def username():
+            if current_user.is_authenticated:
+                return current_user.name
+            return u'Anonymous', 401
+
+        @self.app.route('/logout')
+        def logout():
+            return unicode(logout_user())
+
+        @self.login_manager.request_loader
+        def load_user_from_request(request):
+            user_id = request.args.get('user_id') or session.get('_user_id')
+            try:
+                user_id = int(float(user_id))
+            except TypeError:
+                pass
+            return USERS.get(user_id)
+
+        # This will help us with the possibility of typoes in the tests. Now
+        # we shouldn't have to check each response to help us set up state
+        # (such as login pages) to make sure it worked: we will always
+        # get an exception raised (rather than return a 404 response)
+        @self.app.errorhandler(404)
+        def handle_404(e):
+            raise e
+
+        unittest.TestCase.setUp(self)
+
+    def test_has_no_user_loader_callback(self):
+        self.assertIsNone(self.login_manager._user_callback)
+
+    def test_request_context_users_are_anonymous(self):
+        with self.app.test_request_context():
+            self.assertTrue(current_user.is_anonymous)
+
+    def test_defaults_anonymous(self):
+        with self.app.test_client() as c:
+            result = c.get('/username')
+            self.assertEqual(result.status_code, 401)
+
+    def test_login_via_request(self):
+        user_id = 2
+        user_name = USERS[user_id].name
+        with self.app.test_client() as c:
+            url = '/username?user_id={user_id}'.format(user_id=user_id)
+            result = c.get(url)
+            self.assertEqual(user_name, result.data.decode('utf-8'))
+
+    def test_login_via_request_uses_cookie_when_already_logged_in(self):
+        user_id = 2
+        user_name = notch.name
+        with self.app.test_client() as c:
+            c.get('/login-notch')
+            url = '/username'
+            result = c.get(url)
+            self.assertEqual(user_name, result.data.decode('utf-8'))
+            url = '/username?user_id={user_id}'.format(user_id=user_id)
+            result = c.get(url)
+            self.assertEqual(u'Steve', result.data.decode('utf-8'))
+
+    def test_login_invalid_user_with_request(self):
+        user_id = 9000
+        with self.app.test_client() as c:
+            url = '/username?user_id={user_id}'.format(user_id=user_id)
+            result = c.get(url)
+            self.assertEqual(result.status_code, 401)
+
+    def test_login_invalid_user_with_request_when_already_logged_in(self):
+        user_id = 9000
+        with self.app.test_client() as c:
+            url = '/login-notch'
+            result = c.get(url)
+            self.assertEqual(u'True', result.data.decode('utf-8'))
+            url = '/username?user_id={user_id}'.format(user_id=user_id)
+            result = c.get(url)
+            self.assertEqual(result.status_code, 401)
+
+    def test_login_user_with_request_does_not_modify_session(self):
+        user_id = 2
+        user_name = USERS[user_id].name
+        with self.app.test_client() as c:
+            url = '/username?user_id={user_id}'.format(user_id=user_id)
+            result = c.get(url)
+            self.assertEqual(user_name, result.data.decode('utf-8'))
+            url = '/username'
+            result = c.get(url)
+            self.assertEqual(u'Anonymous', result.data.decode('utf-8'))
+
+
 class TestLoginUrlGeneration(unittest.TestCase):
     def setUp(self):
         self.app = Flask(__name__)
@@ -1316,6 +1470,25 @@ class CookieEncodingTestCase(unittest.TestCase):
             self.assertIsNone(decode_cookie(u'Foo|BAD_BASH'))
             self.assertIsNone(decode_cookie(u'no bar'))
 
+    def test_cookie_encoding_with_key(self):
+        app = Flask(__name__)
+        app.config['SECRET_KEY'] = 'not-used'
+        key = 'deterministic'
+
+        # COOKIE = u'1|7d276051c1eec578ed86f6b8478f7f7d803a7970'
+
+        # Due to the restriction of 80 chars I have to break up the hash in two
+        h1 = u'0e9e6e9855fbe6df7906ec4737578a1d491b38d3fd5246c1561016e189d6516'
+        h2 = u'043286501ca43257c938e60aad77acec5ce916b94ca9d00c0bb6f9883ae4b82'
+        h3 = u'ae'
+        COOKIE = u'1|' + h1 + h2 + h3
+
+        with app.test_request_context():
+            self.assertEqual(COOKIE, encode_cookie(u'1', key=key))
+            self.assertEqual(u'1', decode_cookie(COOKIE, key=key))
+            self.assertIsNone(decode_cookie(u'Foo|BAD_BASH', key=key))
+            self.assertIsNone(decode_cookie(u'no bar', key=key))
+
 
 class SecretKeyTestCase(unittest.TestCase):
     def setUp(self):
@@ -1397,7 +1570,7 @@ class UnicodeCookieUserIDTestCase(unittest.TestCase):
         self.app.config['REMEMBER_COOKIE_NAME'] = self.remember_cookie_name
         self.login_manager = LoginManager()
         self.login_manager.init_app(self.app)
-        self.login_manager._login_disabled = False
+        self.app.config['LOGIN_DISABLED'] = False
 
         @self.app.route('/')
         def index():
@@ -1465,7 +1638,7 @@ class StrictHostForRedirectsTestCase(unittest.TestCase):
         self.app.config['REMEMBER_COOKIE_NAME'] = self.remember_cookie_name
         self.login_manager = LoginManager()
         self.login_manager.init_app(self.app)
-        self.login_manager._login_disabled = False
+        self.app.config['LOGIN_DISABLED'] = False
 
         @self.app.route('/secret')
         def secret():
@@ -1517,9 +1690,12 @@ class StrictHostForRedirectsTestCase(unittest.TestCase):
             self.assertEqual(result.location,
                              'http://good.com/login?next=%2Fsecret')
 
+    @unittest.skipIf(Version(werkzeug_version) < Version('0.15', partial=True),
+                     "ProxyFix moved to werkzeug.middleware.proxy_fix in 0.15")
     def test_unauthorized_uses_host_from_x_forwarded_for_header(self):
         self.login_manager.login_view = 'login'
         self.app.config['FORCE_HOST_FOR_REDIRECTS'] = None
+        self.app.wsgi_app = ProxyFix(self.app.wsgi_app, x_host=1)
 
         @self.app.route('/login')
         def login():
@@ -1554,3 +1730,63 @@ class StrictHostForRedirectsTestCase(unittest.TestCase):
             self.assertEqual(result.status_code, 302)
             self.assertEqual(result.location,
                              'http://good.com/login?next=%2Fsecret')
+
+
+class CustomTestClientTestCase(unittest.TestCase):
+    def setUp(self):
+        self.app = Flask(__name__)
+        self.app.config['SECRET_KEY'] = 'deterministic'
+        self.app.config['SESSION_PROTECTION'] = None
+        self.remember_cookie_name = 'remember'
+        self.app.config['REMEMBER_COOKIE_NAME'] = self.remember_cookie_name
+        self.login_manager = LoginManager()
+        self.login_manager.init_app(self.app)
+        self.app.config['LOGIN_DISABLED'] = False
+        self.app.test_client_class = FlaskLoginClient
+
+        @self.app.route('/')
+        def index():
+            return u'Welcome!'
+
+        @self.app.route('/username')
+        def username():
+            if current_user.is_authenticated:
+                return current_user.name
+            return u'Anonymous'
+
+        @self.app.route('/is-fresh')
+        def is_fresh():
+            return unicode(login_fresh())
+
+        @self.login_manager.user_loader
+        def load_user(user_id):
+            return USERS[int(user_id)]
+
+        # This will help us with the possibility of typoes in the tests. Now
+        # we shouldn't have to check each response to help us set up state
+        # (such as login pages) to make sure it worked: we will always
+        # get an exception raised (rather than return a 404 response)
+        @self.app.errorhandler(404)
+        def handle_404(e):
+            raise e
+
+        unittest.TestCase.setUp(self)
+
+    def test_no_args_to_test_client(self):
+        with self.app.test_client() as c:
+            result = c.get('/username')
+            self.assertEqual(u'Anonymous', result.data.decode('utf-8'))
+
+    def test_user_arg_to_test_client(self):
+        with self.app.test_client(user=notch) as c:
+            username = c.get('/username')
+            self.assertEqual(u'Notch', username.data.decode('utf-8'))
+            is_fresh = c.get('/is-fresh')
+            self.assertEqual(u'True', is_fresh.data.decode('utf-8'))
+
+    def test_fresh_login_arg_to_test_client(self):
+        with self.app.test_client(user=creeper, fresh_login=False) as c:
+            username = c.get('/username')
+            self.assertEqual(u'Creeper', username.data.decode('utf-8'))
+            is_fresh = c.get('/is-fresh')
+            self.assertEqual(u'False', is_fresh.data.decode('utf-8'))
